#include <ConnectionManager.h>
#include <Destination.h>

#include <sstream>

#include <logging.h>

#include <soci/mysql/soci-mysql.h>

#include <mysql.h>

namespace FitsImporter_ns
{

//=============================================================================
//	ConnectionManager::ConnectionManager()
//=============================================================================
ConnectionManager::ConnectionManager(Tango::DeviceImpl* deviceImpl_p,
	Configuration::SP configuration_sp, Instrument::SPVector instrument_spvector)
	throw(std::runtime_error) : Tango::LogAdapter(deviceImpl_p)
{
    DEBUG_STREAM << "ConnectionManager::ConnectionManager()" << endl;

	unsigned int connectionNumber = configuration_sp->getConnectionNumber();

	if(connectionNumber < 1)
		throw std::runtime_error("Connection number property must be greater than 0");

	Instrument::SPVector::iterator it;
	for(it=instrument_spvector.begin(); it!=instrument_spvector.end(); it++)
	{
        DEBUG_STREAM << "ConnectionManager::ConnectionManager() instrument "
            << (*it)->getName().c_str() << endl;

		Destination::SP destination_sp = (*it)->getDestination();

		std::string host = destination_sp->getHost();
		int port = destination_sp->getPort();
		std::string user = destination_sp->getUser();
		std::string password = destination_sp->getPassword();
		std::string schema = destination_sp->getSchema();

		PoolID poolID(host, port, user);

		if(m_pool.find(poolID) == m_pool.end())
		{
			PoolSP pool_sp( new soci::connection_pool(connectionNumber) );

			for(unsigned int i=0; i<connectionNumber; i++)
			{
				std::stringstream connection;
				connection << " host=" << host;
				connection << " port=" << port;
				connection << " user=" << user;
				connection << " password=" << password;
				connection << " dbname=" << schema;

                #ifdef VERBOSE_DEBUG
                    INFO_STREAM << "CONNECTION: " << connection.str() << endl;
                #endif

				pool_sp->at(i).open(soci::mysql, connection.str());

                soci::mysql_session_backend* backend =
                    static_cast<soci::mysql_session_backend*>(
                    pool_sp->at(i).get_backend());

                MYSQL* mysql = backend->conn_;

                bool reconnect = true;
                mysql_options(mysql, MYSQL_OPT_RECONNECT, &reconnect);
			}

			m_pool.insert( std::pair<PoolID, PoolSP>(poolID, pool_sp) );

            INFO_STREAM << "ConnectionManager::ConnectionManager() connection"
                << " host: " << host.c_str() << " port: " << port
                << " user: " << user.c_str() << " added" << endl;
		}
		else
		{
            INFO_STREAM << "ConnectionManager::ConnectionManager() connection"
                << " host: " << host.c_str() << " port: " << port
                << " user: " << user.c_str() << " skipped" << endl;
		}
	}
}

//=============================================================================
//	ConnectionManager::ConnectionManager()
//=============================================================================
ConnectionManager::~ConnectionManager()
{
	DEBUG_STREAM << "ConnectionManager::~ConnectionManager()" << endl;
}

//=============================================================================
//	ConnectionManager::create()
//=============================================================================
ConnectionManager::SP ConnectionManager::create(Tango::DeviceImpl* deviceImpl_p,
	Configuration::SP configuration_sp, Instrument::SPVector instrument_spvector)
	throw(std::runtime_error)
{
	ConnectionManager::SP c_sp(new ConnectionManager(deviceImpl_p,
        configuration_sp, instrument_spvector), ConnectionManager::Deleter());

	return c_sp;
}

//=============================================================================
//	ConnectionManager::getSession()
//=============================================================================
ConnectionManager::SessionSP ConnectionManager::getSession(
	Destination::SP destination_sp) throw(std::runtime_error)
{
	DEBUG_STREAM << "ConnectionManager::getSession()" << endl;

	std::string host = destination_sp->getHost();
	int port = destination_sp->getPort();
	std::string user = destination_sp->getUser();

	PoolID poolID(host, port, user);

	std::map<PoolID, PoolSP>::iterator it;
	if((it = m_pool.find( poolID )) == m_pool.end())
	{
		std::stringstream error_str;
		error_str << "Connection not found for host " << host
			<< ":" << port << " user: " << user;
		throw std::runtime_error(error_str.str());
	}

	SessionSP session_sp( new soci::session( *(it->second) ) );

    //Check if connection is still alive
    if(session_sp->get_backend() == NULL)
        session_sp->reconnect();

	return session_sp;
}

}   //namespace
