#include <DMDBImporter.h>

#include <soci.h>
#include <soci/rowset.h>
#include <soci/use.h>
#include <soci/mysql/soci-mysql.h>
#include <soci/boost-optional.h>
#include <soci/boost-tuple.h>

namespace FitsImporter_ns
{

//==============================================================================
//      DMDBImporter::DMDBImporter()
//==============================================================================
DMDBImporter::DMDBImporter(Tango::DeviceImpl* deviceImpl_p,
    std::string host, unsigned short port, std::string user, std::string password,
    std::string schema, std::string instrumentTable, std::string destinationTable,
    std::string mappingTable) : Tango::LogAdapter(deviceImpl_p), m_host(host),
    m_port(port), m_user(user), m_password(password), m_schema(schema),
    m_instrumentTable(instrumentTable), m_destinationTable(destinationTable),
    m_mappingTable(mappingTable)
{
    DEBUG_STREAM << "DMDBImporter::DMDBImporter()" << endl;

    m_session_sp.reset(new soci::session);
}

//==============================================================================
//      DMDBImporter::~DMDBImporter()
//==============================================================================
DMDBImporter::~DMDBImporter()
{
    DEBUG_STREAM << "DMDBImporter::~DMDBImporter()" << endl;

    m_session_sp->close();
}

//==============================================================================
//      DMDBImporter::create()
//==============================================================================
DMDBImporter::SP DMDBImporter::create(Tango::DeviceImpl* deviceImpl_p,
    std::string host, unsigned short port, std::string user, std::string password,
    std::string schema, std::string instrumentTable, std::string destinationTable,
    std::string mappingTable)
{
    DMDBImporter::SP d_sp(new DMDBImporter(deviceImpl_p, host, port,
        user, password, schema, instrumentTable, destinationTable, mappingTable),
        DMDBImporter::Deleter());

    return d_sp;
}

//==============================================================================
//      DMDBImporter::connect()
//==============================================================================
void DMDBImporter::connect() throw(soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::connect()" << endl;

    std::stringstream connection;
    connection << " host=" << m_host;
    connection << " port=" << m_port;
    connection << " user=" << m_user;
    connection << " password=" << m_password;
    connection << " dbname=" << m_schema;

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "CONNECTION: " << connection.str() << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

    m_session_sp->open(soci::mysql, connection.str());
}

//==============================================================================
//      DMDBImporter::isConnected()
//==============================================================================
bool DMDBImporter::isConnected()
{
    DEBUG_STREAM << "DMDBImporter::isConnect()" << endl;

    if(m_session_sp->get_backend() == NULL)
        return false;
    else
        return true;
}

//==============================================================================
//      DMDBImporter::disconnect()
//==============================================================================
void DMDBImporter::disconnect()
{
    DEBUG_STREAM << "DMDBImporter::disconnect()" << endl;

    m_session_sp->close();
}

//==============================================================================
//      DMDBImporter::retrieveInstrumentList()
//==============================================================================
Instrument::SPVector DMDBImporter::retrieveInstrumentList(std::vector<std::string>&
      instrumentList) throw(std::invalid_argument, soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::retrieveInstrumentList()" << endl;

    if(instrumentList.empty())
        throw std::invalid_argument("Instrument list is empty");

    Instrument::SPVector instrument_spvector;

    std::vector<std::string>::iterator it;
    for(it=instrumentList.begin(); it!=instrumentList.end(); ++it)
    {
        if(it->empty())
            throw std::invalid_argument("Instrument list contains empty element");

        instrument_spvector.push_back(getInstrument(*it));
    }

    return instrument_spvector;
}

//==============================================================================
//      DMDBImporter::retrieveInstrument()
//==============================================================================
Instrument::SP DMDBImporter::retrieveInstrument(std::string instrument)
    throw(std::invalid_argument, soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::retrieveInstrument()" << endl;

    if(instrument.empty())
        throw std::invalid_argument("Instrument is empty");

    return getInstrument(instrument);
}

//==============================================================================
//      DMDBImporter::getInstrument()
//==============================================================================
Instrument::SP DMDBImporter::getInstrument(std::string instrumentName)
        throw(soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::getInstrument()" << endl;

    soci::rowset< InstrumentTuple > rows = (m_session_sp->prepare <<
        "SELECT id, name, fits_key_hdu, fits_key, fits_value, fits_date_hdu, fits_date, dest_id "
        "FROM " << m_instrumentTable << " WHERE name=:name",
        soci::use(instrumentName, "name"));

    InstrumentTupleVector instrumentTuple_vector;
    std::copy(rows.begin(), rows.end(), std::back_inserter(instrumentTuple_vector));

    if(instrumentTuple_vector.empty())
        throw soci::soci_error("No tuple for instrument " + instrumentName);

    if(instrumentTuple_vector.size() > 1)
        throw soci::soci_error("Too tuples for instrument " + instrumentName);

    InstrumentTuple instrumentTuple = instrumentTuple_vector.at(0);

    if(!instrumentTuple.get<0>())
        throw soci::soci_error(" not found");
    int instId = instrumentTuple.get<0>().get();

    if(!instrumentTuple.get<1>())
        throw soci::soci_error("Name not found for id " + instId);
    std::string name = instrumentTuple.get<1>().get();

    if(!instrumentTuple.get<2>())
        throw soci::soci_error("FitsKeyHdu not found for id " + instId);
    int fitsKeyHdu = instrumentTuple.get<2>().get();

    if(!instrumentTuple.get<3>())
        throw soci::soci_error("FitsKey not found for id " + instId);
    std::string fitsKey = instrumentTuple.get<3>().get();

    if(!instrumentTuple.get<4>())
        throw soci::soci_error("FitsValue not found for id " + instId);
    std::string fitsValue = instrumentTuple.get<4>().get();

    if(!instrumentTuple.get<5>())
        throw soci::soci_error("FitsDateHdu not found for id " + instId);
    int fitsDateHdu = instrumentTuple.get<5>().get();

    if(!instrumentTuple.get<6>())
        throw soci::soci_error("FitsDate not found for id " + instId);
    std::string fitsDate = instrumentTuple.get<6>().get();

    if(!instrumentTuple.get<7>())
        throw soci::soci_error("DestId not found for id " + instId);
    int destinationId = instrumentTuple.get<7>().get();

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "INSTRUMENT: " << instId << " " << name << " " << fitsKeyHdu
        << " " << fitsKey << " " << fitsValue << " "
	<< fitsDateHdu << " " << fitsDate << " " << destinationId << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

    Destination::SP destination_sp = getDestination(destinationId);

    Instrument::SP instrument_sp = Instrument::create(name, fitsKeyHdu, fitsKey, fitsValue,
        fitsDateHdu, fitsDate, destination_sp);

    return instrument_sp;
}

//==============================================================================
//      DMDBImporter::getDestination()
//==============================================================================
Destination::SP DMDBImporter::getDestination(int destinationId)
    throw(soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::getDestination()" << endl;

    soci::rowset< DestinationTuple > rows = (m_session_sp->prepare <<
        "SELECT id, host, port, user, password, schema_name, table_name, "
        "storage_path, dir_name FROM " << m_destinationTable << " WHERE id=:id",
        soci::use(destinationId,"id"));

    DestinationTupleVector destinationTuple_vector;
    std::copy(rows.begin(), rows.end(), std::back_inserter(destinationTuple_vector));

    if(destinationTuple_vector.empty())
            throw soci::soci_error("No tuple for destination id " + destinationId);

    if(destinationTuple_vector.size() > 1)
            throw soci::soci_error("Too tuples for destination id " + destinationId);

    DestinationTuple destinationTuple = destinationTuple_vector.at(0);

    if(!destinationTuple.get<1>())
            throw soci::soci_error("Host not found for id " + destinationId);
    std::string host = destinationTuple.get<1>().get();

    if(!destinationTuple.get<2>())
            throw soci::soci_error("Port not found for id " + destinationId);
    int port = destinationTuple.get<2>().get();

    if(!destinationTuple.get<3>())
            throw soci::soci_error("User not found for id " + destinationId);
    std::string user = destinationTuple.get<3>().get();

    if(!destinationTuple.get<4>())
            throw soci::soci_error("Password not found for id " + destinationId);
    std::string password = destinationTuple.get<4>().get();

    if(!destinationTuple.get<5>())
            throw soci::soci_error("Schema not found for id " + destinationId);
    std::string schema = destinationTuple.get<5>().get();

    if(!destinationTuple.get<6>())
            throw soci::soci_error("Table not found for id " + destinationId);
    std::string table = destinationTuple.get<6>().get();

    if(!destinationTuple.get<7>())
            throw soci::soci_error("Storage not found for id " + destinationId);
    std::string storage = destinationTuple.get<7>().get();

    if(!destinationTuple.get<8>())
            throw soci::soci_error("Directory not found for id " + destinationId);
    std::string directory = destinationTuple.get<8>().get();

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "DESTINATION: " << destinationId << " " << host << " " << port
        << " " << user << " " << password << " " << schema << " " << table
        << " " << storage << " " << directory << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

    Mapping::SPVector mapping_spvector = getMapping(destinationId);

    Destination::SP destination_sp( Destination::create(host, port, user, password,
            schema, table, storage, directory, mapping_spvector) );

    return destination_sp;
}

//==============================================================================
//      DMDBImporter::getMapping()
//==============================================================================
Mapping::SPVector DMDBImporter::getMapping(int destinationId)
    throw(soci::soci_error)
{
    DEBUG_STREAM << "DMDBImporter::getMapping()" << endl;

    soci::rowset< MappingTuple > rows = (m_session_sp->prepare << "SELECT id,"
    "column_name, column_type, fits_key_hdu, fits_key_pri, fits_key_sec, mandatory "
    "FROM " << m_mappingTable << " WHERE dest_id=:id", soci::use(destinationId,"id"));

    MappingTupleVector mappingTupleList;
    std::copy(rows.begin(), rows.end(), std::back_inserter(mappingTupleList));

    Mapping::SPVector mapping_spvector;

    MappingTupleVector::iterator it;
    for(it=mappingTupleList.begin(); it!=mappingTupleList.end(); it++)
    {
        if(!it->get<0>())
                throw soci::soci_error("Id not found");
        int mapId = it->get<0>().get();

        if(!it->get<1>())
                throw soci::soci_error("ColumnName not found for id " + mapId);
        std::string columnName = it->get<1>().get();

        if(!it->get<2>())
                throw soci::soci_error("ColumnType not found for id " + mapId);
        std::string columnType = it->get<2>().get();

        if(!it->get<3>())
                throw soci::soci_error("FitsHDU not found for id " + mapId);
        int fitsHDU = it->get<3>().get();

        if(!it->get<4>())
                throw soci::soci_error("FitsPrimary not found for id " + mapId);
        std::string fitsPrimary = it->get<4>().get();

        if(!it->get<5>())
                throw soci::soci_error("FitsSecondary not found for id " + mapId);
        std::string fitsSecondary = it->get<5>().get();

        if(!it->get<6>())
                throw soci::soci_error("Mandatory not found for id " + mapId);
        int mandatory = it->get<6>().get();

#ifdef VERBOSE_DEBUG
        INFO_STREAM << "MAPPING: " << mapId << " " << columnName << " "
            << columnType << " " << fitsHDU << " " << fitsPrimary << " "
            << fitsSecondary << " " << mandatory << endl;
        INFO_STREAM << "-------------------------------------------------" << endl;
#endif

        mapping_spvector.push_back( Mapping::create(columnName, columnType,
            fitsHDU, fitsPrimary, fitsSecondary, mandatory) );
    }

    return mapping_spvector;
}

}   //namespace
