#include <DMDBVerifier.h>

#include <algorithm>

#include <boost/bind.hpp>

#include <soci.h>
#include <soci/rowset.h>
#include <soci/use.h>
#include <soci/boost-optional.h>
#include <soci/boost-tuple.h>

namespace FitsImporter_ns
{

//==============================================================================
//      DMDBVerifier::DMDBVerifier()
//==============================================================================
DMDBVerifier::DMDBVerifier(Tango::DeviceImpl* deviceImpl_p,
    ConnectionManager::SP connectionManager_sp) : Tango::LogAdapter(deviceImpl_p),
    m_connectionManager_sp(connectionManager_sp)
{
    DEBUG_STREAM << "DMDBVerifier::DMDBVerifier()" << endl;
}

//==============================================================================
//      DMDBVerifier::~DMDBVerifier()
//==============================================================================
DMDBVerifier::~DMDBVerifier()
{
    DEBUG_STREAM << "DMDBVerifier::~DMDBVerifier()" << endl;
}

//==============================================================================
//      DMDBVerifier::create()
//==============================================================================
DMDBVerifier::SP DMDBVerifier::create(Tango::DeviceImpl* deviceImpl_p,
    ConnectionManager::SP connectionManager_sp)
{
    DMDBVerifier::SP d_sp(new DMDBVerifier(deviceImpl_p, connectionManager_sp),
        DMDBVerifier::Deleter());

    return d_sp;
}

//==============================================================================
//      DMDBVerifier::testInstrumentListMapping()
//==============================================================================
void DMDBVerifier::testInstrumentListMapping(Instrument::SPVector instrument_spvector)
    throw(soci::soci_error, std::runtime_error)
{
    DEBUG_STREAM << "DMDBVerifier::testInstrumentListMapping()" << endl;

    Instrument::SPVector::iterator it;
    for(it=instrument_spvector.begin(); it!=instrument_spvector.end(); ++it)
        testInstrumentMapping(*it);
}

//==============================================================================
//      DMDBVerifier::testInstrumentMapping()
//==============================================================================
void DMDBVerifier::testInstrumentMapping(Instrument::SP instrument_sp)
    throw(soci::soci_error, std::runtime_error)
{
    DEBUG_STREAM << "DMDBVerifier::testInstrumentMapping()" << endl;

    Destination::SP destination_sp = instrument_sp->getDestination();

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "TABLE: " << destination_sp->getHost() << " "
        << destination_sp->getPort() << " " << destination_sp->getSchema()
         << " " << destination_sp->getTable() << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

    Mapping::SPVector mapping_spvector = destination_sp->getMappingSPVector();

	ConnectionManager::SessionSP session_sp =
		m_connectionManager_sp->getSession(destination_sp);

    soci::rowset< ColumnTuple > rows = (session_sp->prepare << "SELECT "
        "column_name, column_type, is_nullable FROM information_schema.columns "
        "WHERE table_schema LIKE :schema AND table_name LIKE :table",
        soci::use(destination_sp->getSchema(), "schema"),
        soci::use(destination_sp->getTable(), "table"));

    ColumnTupleVector columnTuple_vector;
    std::copy(rows.begin(), rows.end(), std::back_inserter(columnTuple_vector));

    try
    {
        //Check nadir mandatory columns in table definition
        testColumnDefinition(columnTuple_vector, "id", "mediumint", true);
        testColumnDefinition(columnTuple_vector,"storage_path","varchar", false);
        testColumnDefinition(columnTuple_vector,"file_path","varchar", false);
        testColumnDefinition(columnTuple_vector,"file_version","smallint", true);
        testColumnDefinition(columnTuple_vector,"file_name","varchar", true);
        testColumnDefinition(columnTuple_vector,"checksum","varchar", false);
        testColumnDefinition(columnTuple_vector,"checksum_gz","varchar", false);
        testColumnDefinition(columnTuple_vector,"update_time","timestamp", true);

        //Check mapping columns in table definition
        Mapping::SPVector::iterator it;
        for(it=mapping_spvector.begin(); it!=mapping_spvector.end(); ++it)
            testColumnDefinition(columnTuple_vector, (*it)->getColumnName(),
                 (*it)->getColumnType(), (*it)->isMandatory());
    }
    catch(std::runtime_error& ex)
    {
        std::stringstream error_stream;
        error_stream << "DMDBVerifier::testInstrumentMapping() "
            << destination_sp->getHost() << ":" << destination_sp->getPort() << " "
            << destination_sp->getSchema() << "." << destination_sp->getTable()
            << " " << ex.what();
        throw std::runtime_error(error_stream.str());
    }
}
//==============================================================================
//      DMDBVerifier::testColumnDefinition()
//==============================================================================
void DMDBVerifier::testColumnDefinition(ColumnTupleVector& columnTuple_vector,
    std::string columnName, std::string columnType, bool mandatory)
    throw(std::runtime_error)
{
    DEBUG_STREAM << "testColumnDefinition::testColumn()" << endl;

    ColumnTupleVector::iterator it;
    it = std::find_if(columnTuple_vector.begin(), columnTuple_vector.end(),
        boost::bind(&DMDBVerifier::isEqual, this, _1, columnName));

    if(it == columnTuple_vector.end())
    {
        std::stringstream error_stream;
        error_stream << "\"" << columnName << "\" column not exists";
        throw std::runtime_error(error_stream.str());
    }

    if(!it->get<1>())
        throw std::runtime_error("empty column_type retrieved from information_schema");

    if(it->get<1>().get().find(columnType) == std::string::npos)
    {
        std::stringstream error_stream;
        error_stream << "\"" << columnName << "\" column with "
            << columnType << "\" type differs form information schema";
        throw std::runtime_error(error_stream.str());
    }

    if(!it->get<2>())
        throw std::runtime_error("empty is_nullable retrieved from information_schema");

    std::string isNullable("NO");

    if(!mandatory)
        isNullable = "YES";

    if(it->get<2>().get().compare(isNullable) != 0)
    {
        std::stringstream error_stream;
        error_stream << "\"" << columnName << "\" column with is nullable "
            << isNullable << "\" type differs form information schema";
        throw std::runtime_error(error_stream.str());
    }

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "COLUMN: " << columnName << " " << columnType
        << " " << isNullable << " -> OK" << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif
}

//==============================================================================
//      DMDBVerifier::isEqual()
//==============================================================================
bool DMDBVerifier::isEqual(ColumnTuple columnTuple, std::string columnName)
    throw(std::runtime_error)
{
    DEBUG_STREAM << "DMDBVerifier::isEqual()" << endl;

    if(!columnTuple.get<0>())
        throw std::runtime_error("empty column_name retrieved from information_schema");

    return columnTuple.get<0>().get().compare(columnName) == 0;
}

}      //namespace