#include <WorkerThread.h>
#include <FitsImporter.h>
#include <md5.h>

#include <cassert>
#include <fstream>
#include <iomanip>
#include <set>

#include <boost/scoped_ptr.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/regex.hpp>
#include <boost/lexical_cast.hpp>

#include <soci/statement.h>
#include <soci/transaction.h>
#include <soci/rowset.h>
#include <soci/boost-optional.h>

namespace FitsImporter_ns
{

//==============================================================================
//	WorkerThread::WorkerThread()
//==============================================================================
WorkerThread::WorkerThread(FitsImporter* fitsImporter_p, EventBuffer::SP eventBuffer_sp,
    ConnectionManager::SP connectionManager_sp, Configuration::SP configuration_sp,
    Instrument::SPVector instrumentList_spvector, Instrument::SP defaultInstrument_sp) :
    Tango::LogAdapter(fitsImporter_p), m_fitsImporter_p(fitsImporter_p),
    m_eventBuffer_sp(eventBuffer_sp), m_connectionManager_sp(connectionManager_sp),
    m_configuration_sp(configuration_sp), m_instrumentList_spvector(instrumentList_spvector),
    m_defaultInstrument_sp(defaultInstrument_sp)
{
	DEBUG_STREAM << "WorkerThread::WorkerThread()" << std::endl;
}

//==============================================================================
//	WorkerThread::~WorkerThread()
//==============================================================================
WorkerThread::~WorkerThread()
{
	DEBUG_STREAM << "WorkerThread::~WorkerThread()" << std::endl;
}

//==============================================================================
//	WorkerThread::workerLoop()
//==============================================================================
void WorkerThread::workerLoop()
{
    DEBUG_STREAM << "WorkerThread::workerLoop(): starting loop" << endl;

	unsigned int waitTime = m_configuration_sp->getWaitTime();

    while(true)
    {
        try
        {
            //Wait new file event to process
            boost::filesystem::path origPath(m_eventBuffer_sp->waitNew());
            std::string fileName = origPath.stem().string();

            //If configured wait after new file event
            boost::this_thread::sleep_for(boost::chrono::milliseconds(waitTime));

            bool completed = false;

            INFO_STREAM << "WorkerThread::workerLoop() processing \""
                << fileName << "\"" << endl;

            try
            {
                //Open fits file
                boost::shared_ptr<CCfits::FITS> fitsFile_sp(
                    new CCfits::FITS(origPath.string(), CCfits::Write));

                //Keep track of already readed HDU to avoid memory leaks
                std::set<int> readHDUSet;

                try
                {
                    //Try to ingest using an instrument in list
                    ingestUsingInstrumentList(origPath, fitsFile_sp, readHDUSet);

                    //Notify archive correctly event
                    completed = true;
                    m_fitsImporter_p->incrementRegularCounter();
                    INFO_STREAM << "WorkerThread::workerLoop() \"" << fileName
                        << "\" archived regularly" << endl;

                    //Remove original file
                    boost::filesystem::remove(origPath);
                }
                catch(CCfits::FitsException& ex)
                {
                    ERROR_STREAM << "WorkerThread::workerLoop() " << ex.message() << endl;
                }
                catch(std::runtime_error& ex)
                {
                    ERROR_STREAM << "WorkerThread::workerLoop() " << ex.what() << endl;
                }

                if(!completed)
                {
                    //Try to ingest using default instrument
                    ingestUsingDefaultInstrument(origPath, fitsFile_sp, readHDUSet);

                    //Notify archiving warning event
                    completed = true;
                    m_fitsImporter_p->incrementWarningCounter();
                    WARN_STREAM << "WorkerThread::workerLoop() \"" << fileName
                        << "\" archived in default instrument" << endl;

                    //Remove original file
                    boost::filesystem::remove(origPath);
                }
            }
            catch(CCfits::FitsException& ex)
            {
                ERROR_STREAM << "WorkerThread::workerLoop() " << ex.message() << endl;
            }
            catch(std::runtime_error& ex)
            {
                ERROR_STREAM << "WorkerThread::workerLoop() " << ex.what() << endl;
            }

            if(!completed)
            {
                //Notify archiving error event
                m_fitsImporter_p->incrementErrorCounter();
                ERROR_STREAM << "WorkerThread::workerLoop() \"" << fileName
                    << "\" not archived" << endl;
            }

            m_eventBuffer_sp->markAsProcessed(origPath);
        }
		catch(boost::thread_interrupted& ex)
		{
            DEBUG_STREAM << "WorkerThread::workerLoop() interrupt" << endl;

			break;
		}
        catch(std::exception& ex)
        {
            ERROR_STREAM << "WorkerThread::workerLoop() " << ex.what() << endl;
        }
		catch(...)
		{
            ERROR_STREAM << "WorkerThread::workerLoop() unknown exception" << endl;
		}
    } //while
}

//==============================================================================
//	WorkerThread::ingestUsingInstrumentList()
//==============================================================================
void WorkerThread::ingestUsingInstrumentList(boost::filesystem::path& origPath,
    boost::shared_ptr<CCfits::FITS> fitsFile_sp, std::set<int>& readHDUSet)
    throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList()" << endl;

    // Read all primary HDU keys and keep track
    // This is useful when all keywords are in the first HDU
    //CCfits::PHDU& phdu = fitsFile_sp->pHDU();
    //phdu.readAllKeys();
    //readHDUSet.insert(0);

    Instrument::SP instrument_sp;
    boost::gregorian::date date;
    bool found = false;

    //Iterate over instrument list
    Instrument::SPVector::iterator it;
    for(it=m_instrumentList_spvector.begin(); it!=m_instrumentList_spvector.end(); ++it)
    {
        int fitsKeyHdu = (*it)->getFitsKeyHdu();
        std::string fitsKey = (*it)->getFitsKey();
        std::string fitsValue = (*it)->getFitsValue();
        int fitsDateHdu = (*it)->getFitsDateHdu();
        std::string fitsDate = (*it)->getFitsDate();

        DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList() instrument: "
            << (*it)->getName() << " keyHdu: " << fitsKeyHdu << " key: " << fitsKey << " value: " << fitsValue
            << " dateHdu: " << fitsDateHdu << " date: " << fitsDate << endl;

        try
        {
            std::string retrievedFitsKeyValue;
            std::string retrievedFitsDateValue;

            //If current fitsKeyHdu has not been already read
            if(readHDUSet.find(fitsKeyHdu) == readHDUSet.end())
            {
                //Read all currnet HDU keys
                if(fitsKeyHdu == 0)
                    fitsFile_sp->pHDU().readAllKeys();
                else
                    fitsFile_sp->extension(fitsKeyHdu).readAllKeys();
                //Keep track of read HDU
                readHDUSet.insert(fitsKeyHdu);
            }

            //If current fitsDateHdu has not been already read
            if(readHDUSet.find(fitsDateHdu) == readHDUSet.end())
            {
                //Read all currnet HDU keys
                if(fitsDateHdu == 0)
                    fitsFile_sp->pHDU().readAllKeys();
                else
                    fitsFile_sp->extension(fitsDateHdu).readAllKeys();
                //Keep track of read HDU
                readHDUSet.insert(fitsDateHdu);
            }

            // Read retrievedFitsKeyValue according to fitsKeyHdu
            if(fitsKeyHdu == 0)
                fitsFile_sp->pHDU().readKey(fitsKey, retrievedFitsKeyValue);
            else
                fitsFile_sp->extension(fitsKeyHdu).readKey(fitsKey, retrievedFitsKeyValue);

            // Read retrievedFitsDateValue according to fitsDateHdu
            if(fitsDateHdu == 0)
                fitsFile_sp->pHDU().readKey(fitsDate, retrievedFitsDateValue);
            else
                fitsFile_sp->extension(fitsDateHdu).readKey(fitsDate, retrievedFitsDateValue);

            // Other ways to get keywords value
            //phdu.readKey(fitsKey, retrievedFitsKeyValue);
            //phdu.readKey(fitsDate, retrievedFitsDateValue);

            if(fitsValue.compare(retrievedFitsKeyValue)==0)
            {
                instrument_sp = *it;
                date = parseDate(retrievedFitsDateValue);
                found = true;
            }
        }
        catch(CCfits::HDU::NoSuchKeyword& ex)
        {
            DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList() "
                << ex.message() << endl;
        }
        catch(boost::bad_lexical_cast& ex)
        {
            DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList() "
                << ex.what() << endl;
        }
        catch(std::runtime_error& ex)
        {
            DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList() "
                << ex.what() << endl;
        }
        catch(CCfits::FITS::NoSuchHDU& ex)
		{
            DEBUG_STREAM << "WorkerThread::ingestUsingInstrumentList() "
                << ex.message() << endl;
        }

        
	} //for

    if(found)
        processFits(origPath, fitsFile_sp, readHDUSet, instrument_sp, date);
    else
        throw std::runtime_error("Instrument not found in instrument list");
}

//==============================================================================
//	WorkerThread::ingestUsingDefaultInstrument()
//===============================================================================
void WorkerThread::ingestUsingDefaultInstrument(boost::filesystem::path& origPath,
    boost::shared_ptr<CCfits::FITS> fitsFile_sp, std::set<int>& readHDUSet)
    throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::ingestUsingDefaultInstrument()" << endl;

    if(!m_defaultInstrument_sp)
        throw std::runtime_error("Default instrument not defined");

    //To use UTC date instead of local time
    //boost::gregorian::date defaultDate = boost::gregorian::day_clock::universal_day();
	boost::gregorian::date defaultDate = boost::gregorian::day_clock::local_day();

    processFits(origPath, fitsFile_sp, readHDUSet, m_defaultInstrument_sp, defaultDate);
}

//==============================================================================
//	WorkerThread::parseDate()
//==============================================================================
boost::gregorian::date WorkerThread::parseDate(std::string dateString)
	throw(boost::bad_lexical_cast, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::parseDate()" << endl;

	std::stringstream dateStream;

	//Regex match date like DD/MM/YYYYY DD-MM-YYYY DD.MM.YYYY
	boost::regex pattern_1("^(0?[1-9]|[1-2][0-9]|3[01])(/|-|\\.)(0?[1-9]|1[012])(/|-|\\.)((19|20)\\d\\d).*$");

	boost::smatch matches_1;
	if(boost::regex_match(dateString, matches_1, pattern_1))
	{
        DEBUG_STREAM << "WorkerThread::parseDate() match date like DD/MM/YYYYY DD-MM-YYYY DD.MM.YYYY" << endl;

		if(matches_1.size() == 7)
			dateStream << matches_1[5] << "/" << matches_1[3] << "/" << matches_1[1];
		else
			throw std::runtime_error("Error parsing date like DD/MM/YYYYY DD-MM-YYYY DD.MM.YYYY");

		return boost::gregorian::date(boost::gregorian::from_string(dateStream.str()));
	}

	//Regex match date like DDMMYYYY
	boost::regex pattern_2("^(0?[1-9]|[1-2][0-9]|3[01])(0?[1-9]|1[012])((19|20)\\d\\d).*$");

	boost::smatch matches_2;
	if(boost::regex_match(dateString, matches_2, pattern_2))
	{
        DEBUG_STREAM << "WorkerThread::parseDate() match date like DDMMYYYY" << endl;

		if(matches_2.size() == 5)
			dateStream << matches_2[3] << "/" << matches_2[2] << "/" << matches_2[1];
		else
			throw std::runtime_error("Error parsing DDMMYYYY");

		return boost::gregorian::date(boost::gregorian::from_string(dateStream.str()));
	}

	//Regex match date like YYYY/MM/DD YYYY-MM-DD YYYY.MM.DD
	boost::regex pattern_3("^((19|20)\\d\\d)(/|-|\\.)(0?[1-9]|1[012])(/|-|\\.)(0+[1-9]|[1-2][0-9]|3[01]).*$");

	boost::smatch matches_3;
	if(boost::regex_match(dateString, matches_3, pattern_3))
	{
        DEBUG_STREAM << "WorkerThread::parseDate() match date like YYYY/MM/DD YYYY-MM-DD YYYY.MM.DD" << endl;

		if(matches_3.size() == 7)
			dateStream << matches_3[1] << "/" << matches_3[4] << "/" << matches_3[6];
		else
			throw std::runtime_error("Error parsing date like YYYY/MM/DD YYYY-MM-DD YYYY.MM.DD");

		return boost::gregorian::date(boost::gregorian::from_string(dateStream.str()));
	}

	//Regex match date like YYYYMMDD
	boost::regex pattern_4("^((19|20)\\d\\d)(0?[1-9]|1[012])(0+[1-9]|[1-2][0-9]|3[01]).*$");

	boost::smatch matches_4;
	if(boost::regex_match(dateString, matches_4, pattern_4))
	{
        DEBUG_STREAM << "WorkerThread::parseDate() match date like YYYYMMDD" << endl;

		if(matches_4.size() == 5)
			dateStream << matches_4[1] << "/" << matches_4[3] << "/" << matches_4[4];
		else
			throw std::runtime_error("Error parsing YYYYMMDD");

		return boost::gregorian::date(boost::gregorian::from_string(dateStream.str()));
	}

	throw std::runtime_error("Date string not match any known format");
}

//==============================================================================
//	WorkerThread::processFits()
//==============================================================================
void WorkerThread::processFits(boost::filesystem::path& origPath,
    boost::shared_ptr<CCfits::FITS> fitsFile_sp, std::set<int>& readHDUSet,
    Instrument::SP instrument_sp, boost::gregorian::date& date)
    throw (CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::processFits()" << endl;

    //	CCfits::PHDU& phdu = fitsFile_sp->pHDU();

	Destination::SP destination_sp = instrument_sp->getDestination();

	//Retrieve a database session
	ConnectionManager::SessionSP session_sp =
		m_connectionManager_sp->getSession(destination_sp);

	//Start a transaction [RAII]
	soci::transaction transaction(*session_sp);

	//Compute checksum
	std::string checksum = computeMd5(origPath);

	//Search if file is already archived
	int duplicateMax = findDuplicateMax(session_sp, destination_sp, origPath, checksum);

	//In case of duplicate, add key IA2VERS with duplicateMax [REMOVED]
//	if(duplicateMax > 0)
//	{
//		phdu.addKey("IA2VERS", duplicateMax, "IA2 version for duplicate file");
//		fitsFile_sp->flush();
//	}

	//Create destination path
	boost::filesystem::path destPath = createDestPath(origPath,
							  destination_sp, date, duplicateMax);


	//Copy file to archive storage
	copyFile(origPath, destPath);
	
	//Compute checksum_gz
	std::string checksum_gz = computeMd5(destPath);

	//Don't add the file to metadata if duplidateMax=-1 (i.e. if it already exists with the same checksum)
	if(duplicateMax!=-1)
	  {
	    //Create metadata query
	    std::string query = composeQuery(destination_sp, date, duplicateMax, checksum, checksum_gz,
					     origPath, fitsFile_sp, readHDUSet);

	    try
	      {
		//Insert metadata into database
		*session_sp << query;
	      }
	    catch(std::runtime_error& ex)
	      {
		//In case of error remove archived file
		boost::filesystem::remove(destPath);
		throw(ex);
	      }
	  }
	
	//Commit Transaction
	transaction.commit();
}

//==============================================================================
//	WorkerThread::findDuplicateMax()
//==============================================================================
int WorkerThread::findDuplicateMax(ConnectionManager::SessionSP session_sp,
	Destination::SP destination_sp, boost::filesystem::path& origPath, std::string checksum)
	throw(std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::findDuplicateMax()" << endl;

	std::stringstream query_checksum;
	query_checksum << "select max(file_version) from " << destination_sp->getSchema()
	      << "." << destination_sp->getTable() << " where file_name = \'"
	      << origPath.filename().string() << ".gz\' and checksum = \'"
	      << checksum << "\'" << std::endl;

	std::stringstream query;
	query << "select max(file_version) from " << destination_sp->getSchema()
	      << "." << destination_sp->getTable() << " where file_name like \'"
	      << origPath.filename().string() << ".gz\'" << std::endl;

    DEBUG_STREAM << "WorkerThread::findDuplicateMax() query " << query_checksum.str() << endl;
    DEBUG_STREAM << "WorkerThread::findDuplicateMax() query " << query.str() << endl;

	boost::optional<int> result_checksum;
	*session_sp << query_checksum.str(), soci::into(result_checksum);

	int duplicateMax = 0;

	if(result_checksum)
	  {
	    duplicateMax=-1;
	    std::string fileName( origPath.filename().string() );
	   
	    WARN_STREAM << "WorkerThread::findDuplicateMax() file " << fileName
			<< " was already ingested -> moved to notnew " << endl;
	  }
	else
	  {
	    boost::optional<int> result;
	    *session_sp << query.str(), soci::into(result);

	    //If result != null
	    if(result)
	      {
		//In case of duplicate increment
		duplicateMax = result.get();
		++duplicateMax;
	      }

	    if(duplicateMax > 0)
	      {
		std::string fileName( origPath.filename().string() );

		WARN_STREAM << "WorkerThread::findDuplicateMax() file " << fileName
			    << " has " << duplicateMax << " duplicates " << endl;
	      }
	  }

	return duplicateMax;
}

//==============================================================================
//	WorkerThread::composeQuery()
//==============================================================================
std::string WorkerThread::composeQuery(Destination::SP destination_sp,
    boost::gregorian::date& date, int duplicateMax, std::string checksum, std::string checksum_gz,
    boost::filesystem::path& origPath, boost::shared_ptr<CCfits::FITS> fitsFile_sp, std::set<int>& readHDUSet)
	throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::composeQuery()" << endl;

	//Insert part of the query
	std::stringstream insertQuery;
	insertQuery << "INSERT INTO " << destination_sp->getSchema() << "."
		<< destination_sp->getTable() << "(";

	//Values part of the query
	std::stringstream valuesQuery;
	valuesQuery << "VALUES(";

	//Append storage path column to query
	insertQuery << destination_sp->getTable() << ".storage_path";
	valuesQuery << "\'" << destination_sp->getStoragePath() << "\'";

	//Append checksum
	insertQuery << "," << destination_sp->getTable() << ".checksum";
	valuesQuery << ",\'" << checksum << "\'";

	//Append checksum_gz
	insertQuery << "," << destination_sp->getTable() << ".checksum_gz";
	valuesQuery << ",\'" << checksum_gz << "\'";

	//Append file path column to query
	insertQuery << "," << destination_sp->getTable() << ".file_path";
	int month = date.month();
	valuesQuery << ",\'/" << date.year();
    valuesQuery << "/" << std::setw(2) << std::setfill('0') << month;
    valuesQuery << "/" << std::setw(2) << std::setfill('0') << date.day();
    valuesQuery << "/" << destination_sp->getDirName() << "\'";

	//Append file version column to query
	insertQuery << "," << destination_sp->getTable()  << ".file_version";
	valuesQuery << "," << duplicateMax;

	//Append file name in column to query
	insertQuery << "," << destination_sp->getTable()  << ".file_name";

	valuesQuery << ",\'" << origPath.filename().string() << ".gz\'";

    //Search key from mapping and fill query
    fillQueryFromFits(insertQuery, valuesQuery, destination_sp, fitsFile_sp, readHDUSet);

	//Merge insert and values queries
	insertQuery << ") " << valuesQuery.rdbuf() << ")";

    DEBUG_STREAM << "WorkerThread::execQuery() query " << insertQuery.str() << endl;

    return insertQuery.str();
}

//==============================================================================
//	WorkerThread::fillQueryFromFits()
//==============================================================================
void WorkerThread::fillQueryFromFits(std::stringstream& insertQuery,
    std::stringstream& valuesQuery, Destination::SP destination_sp,
    boost::shared_ptr<CCfits::FITS> fitsFile_sp, std::set<int>& readHDUSet)
    throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::fillQueryFromFits()" << endl;

	//Get mapping for a specific instrument
	Mapping::SPVector mapping_spvector = destination_sp->getMappingSPVector();

	//Loop mapping to append metadata columns to query
	Mapping::SPVector::iterator it;
	for(it=mapping_spvector.begin(); it!=mapping_spvector.end(); it++)
	{
        int fitsHDU = (*it)->getFitsHDU();

        //If current HDU has not been already read
        if(readHDUSet.find(fitsHDU) == readHDUSet.end())
        {
            //Read all currnet HDU keys
            if(fitsHDU == 0)
                fitsFile_sp->pHDU().readAllKeys();
            else
                fitsFile_sp->extension(fitsHDU).readAllKeys();

            //Keep track of read HDU
            readHDUSet.insert(fitsHDU);
        }

        std::string value;

        try
        {
            try
            {
                if(fitsHDU == 0)
                {
                    //Read primary key from PHDU
                    value = readKeyFromPHDU(fitsFile_sp->pHDU(),
                        (*it)->getFitsPrimary(), (*it)->getColumnType());
                }
                else
                {
                    //Read primary key from extension HDU
                    value = readKeyFromHDU(fitsFile_sp->extension(fitsHDU),
                        (*it)->getFitsPrimary(), (*it)->getColumnType());
                }

                insertQuery << "," << destination_sp->getTable() << "."
                        << (*it)->getColumnName();
                valuesQuery << "," << value;
            }
            catch(CCfits::HDU::NoSuchKeyword& ex)
            {
                WARN_STREAM << "WorkerThread::fillQueryFromFits() " << ex.message() << endl;

                try
                {
                    if(fitsHDU == 0)
                    {
                        //Read secondary key from PHDU
                        value = readKeyFromPHDU(fitsFile_sp->pHDU(),
                            (*it)->getFitsSecondary(), (*it)->getColumnType());
                    }
                    else
                    {
                        //Read secondary key from extension HDU
                        value = readKeyFromHDU(fitsFile_sp->extension(fitsHDU),
                            (*it)->getFitsSecondary(), (*it)->getColumnType());
                    }

                    insertQuery << "," << destination_sp->getTable() << "."
                        << (*it)->getColumnName();
                    valuesQuery << "," << value;
                }
                catch(CCfits::HDU::NoSuchKeyword& ex)
                {
                    WARN_STREAM << "WorkerThread::fillQueryFromFits() " << ex.message() << endl;

                    //If key is mandatory throw and exception
                    if((*it)->isMandatory())
                        throw ex;
                }
            }
		}
		catch(CCfits::FITS::NoSuchHDU& ex)
		{
            WARN_STREAM << "WorkerThread::fillQueryFromFits() " << ex.message() << endl;

            //If key is mandatory throw and exception
            if((*it)->isMandatory())
                throw ex;
        }
	}
}

//==============================================================================
//	WorkerThread::readKeyFromPHDU()
//==============================================================================
std::string WorkerThread::readKeyFromPHDU(CCfits::PHDU& phdu, std::string keyName,
    std::string keyType) throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::readKeyFromPHDU()" << endl;

    std::stringstream resultStream;

    boost::regex pattern("^(NAXIS)([0-9]+)$", boost::regex::icase);
    boost::smatch matches;

    if(keyName.compare("NAXIS")==0 || keyName.compare("naxis")==0)
    {
        resultStream << phdu.axes();
    }
    else if(boost::regex_match(keyName, matches, pattern))
    {
        long axisMax = phdu.axes();

        try
        {
            long axisReq = boost::lexical_cast<long>(matches[2]) - 1;

            if(axisReq<0 || axisReq>=axisMax)
                throw CCfits::HDU::NoSuchKeyword("Invalid naxis number");

            resultStream << phdu.axis(axisReq);
        }
        catch(boost::bad_lexical_cast& ex)
        {
            throw std::runtime_error("Exception parsing naxis number");
        }
    }
    else if(keyName.compare("BITPIX")==0 || keyName.compare("bitpix")==0)
    {
        resultStream << phdu.bitpix();
    }
    else if(keyName.compare("BSCALE")==0 || keyName.compare("bscale")==0)
    {
        resultStream << phdu.scale();
    }
    else if(keyName.compare("BZERO")==0 || keyName.compare("bzero")==0)
    {
        resultStream << phdu.zero();
    }
    else if(keyName.compare("EXTEND")==0 || keyName.compare("extend")==0)
    {
        resultStream << phdu.extend();
    }
    else if(keyType.compare("varchar")==0 || keyType.compare("char")==0)
    {
        resultStream << "\'" << readKey(phdu, keyName, keyType) << "\'";
    }
    else if(keyType.compare("date")==0 || keyType.compare("datetime")==0 ||
        keyType.compare("time")==0 || keyType.compare("timestamp")==0)
    {
        resultStream << "\'" << readKey(phdu, keyName, keyType) << "\'";
    }
    else
    {
        resultStream << readKey(phdu, keyName, keyType);
    }

    return resultStream.str();
}

//==============================================================================
//	WorkerThread::readKeyFromHDU()
//==============================================================================
std::string WorkerThread::readKeyFromHDU(CCfits::HDU& hdu, std::string keyName,
    std::string keyType) throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::readKeyFromHDU()" << endl;

    std::stringstream resultStream;

    boost::regex pattern("^(NAXIS)([0-9]+)$", boost::regex::icase);
    boost::smatch matches;

    if(keyName.compare("NAXIS")==0 || keyName.compare("naxis")==0)
    {
        resultStream << hdu.axes();
    }
    else if(boost::regex_match(keyName, matches, pattern))
    {
        long axisMax = hdu.axes();

        try
        {
            long axisReq = boost::lexical_cast<long>(matches[2]) - 1;

            if(axisReq<0 || axisReq>=axisMax)
                throw std::runtime_error("Invalid naxis number");

            resultStream << hdu.axis(axisReq);
        }
        catch(boost::bad_lexical_cast& ex)
        {
            throw std::runtime_error("Exception parsing naxis number");
        }
    }
    else if(keyName.compare("BITPIX")==0 || keyName.compare("bitpix")==0)
    {
        resultStream << hdu.bitpix();
    }
    else if(keyName.compare("BSCALE")==0 || keyName.compare("bscale")==0)
    {
        resultStream << hdu.scale();
    }
    else if(keyName.compare("BZERO")==0 || keyName.compare("bzero")==0)
    {
        resultStream << hdu.zero();
    }
    else if(keyType.compare("varchar")==0 || keyType.compare("char")==0)
    {
        resultStream << "\'" << readKey(hdu, keyName, keyType) << "\'";
    }
    else if(keyType.compare("date")==0 || keyType.compare("datetime")==0 ||
        keyType.compare("time")==0 || keyType.compare("timestamp")==0)
    {
        resultStream << "\'" << readKey(hdu, keyName, keyType) << "\'";
    }
    else
    {
        resultStream << readKey(hdu, keyName, keyType);
    }

    return resultStream.str();
}

//==============================================================================
//	WorkerThread::readKey()
//==============================================================================
std::string WorkerThread::readKey(CCfits::HDU& hdu, std::string keyName,
    std::string keyType) throw(CCfits::FitsException, std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::readKey()" << endl;

	int fitsKeyType = hdu.keyWord(keyName).keytype();

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "KEY: " << keyName << " " << fitsKeyType << " -> " << keyType << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

    try
    {
        switch(fitsKeyType)
        {
            case TSTRING:
            {
                std::string value;
                hdu.readKey(keyName, value);
                return value;
            }
            case TLOGICAL:
            {
                bool value;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TINT: case TSHORT: case TSBYTE:
            {
                int value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TUINT: case TUSHORT: case TBIT: case TBYTE:
            {
                unsigned int value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TULONG:
            {
                unsigned long value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TLONG:
            {
                long value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TLONGLONG:
            {
                long long value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TDOUBLE: case TFLOAT:
            {
                double value = 0;
                hdu.readKey(keyName, value);
                return boost::lexical_cast<std::string>(value);
            }
            case TCOMPLEX: case TDBLCOMPLEX:
            {
                //TODO: [WARNING] complex case
                std::runtime_error(" Complex fits type not yet implemented");
            }
            default:
                throw std::runtime_error("Unknown fits type");
        }   //switch
    }
    catch(boost::bad_lexical_cast& ex)
    {
        throw std::runtime_error(ex.what());
    }
}

//==============================================================================
//	WorkerThread::createDestPath()
//==============================================================================
boost::filesystem::path WorkerThread::createDestPath(boost::filesystem::path& origPath,
	Destination::SP destination_sp, boost::gregorian::date& date,
	int duplicateMax) throw(std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::createDestPath()" << endl;

    //Destination path = storage path + directory name
    // + date path + duplicate max + file name
    boost::filesystem::path destPath(destination_sp->getStoragePath());
    
    //If duplicateMax is -1 use the today date
    //and add notnew to destPath
    if(duplicateMax == -1)
      {
	date = boost::gregorian::day_clock::local_day();
	destPath /= "/notnew";
      }
    
    //Create directory path, date and version part of destination path
    std::stringstream relPathStream;
    int month = date.month();
    relPathStream << "/" << date.year();
    relPathStream << "/" << std::setw(2) << std::setfill('0') << month ;
    relPathStream << "/" << std::setw(2) << std::setfill('0') << date.day();
    
    //If duplicateMax is -1 do not add getDirName and duplicateMax to destination path
    if(duplicateMax != -1)
      relPathStream << "/" << destination_sp->getDirName() << "/" << duplicateMax;
    boost::filesystem::path relPath( relPathStream.str() );
    
    //Append date path to full destination path
    destPath /= relPath;
    
    //If destination path not exists, create it
    if(!boost::filesystem::exists(destPath))
      boost::filesystem::create_directories(destPath);
    
    //Check if destination path is a directory
    if(!boost::filesystem::is_directory(destPath))
      throw std::runtime_error( "Destination path \"" +
				destPath.string() + "\" is not a directory");
    
    std::stringstream destFileName;
    destFileName << origPath.filename().string() << ".gz";
    boost::filesystem::path destFile(destFileName.str());
    
    //Append file name to destination path
    destPath /= destFile;
    
    return destPath;
}

//==============================================================================
//	WorkerThread::copyFile()
//==============================================================================
void WorkerThread::copyFile(boost::filesystem::path& origPath,
	boost::filesystem::path& destPath) throw(std::runtime_error)
{
    DEBUG_STREAM << "WorkerThread::copyFile()" << endl;

	//Is origin path a regular file
	if(!boost::filesystem::is_regular_file(origPath))
		throw std::runtime_error( "Origin path \"" +
			origPath.string() + "\" is not a regular file");

#ifdef VERBOSE_DEBUG
    INFO_STREAM << "FROM: " << origPath.string() << " TO: " << destPath.string() << endl;
    INFO_STREAM << "-------------------------------------------------" << endl;
#endif

	//Create input and output stream
	std::ifstream inStream(origPath.string().c_str(), std::ios_base::binary);
	std::ofstream outStream(destPath.string().c_str(), std::ios_base::binary);

	//Add gzip compressor on output stream
	boost::iostreams::filtering_streambuf<boost::iostreams::output> out;
	out.push(boost::iostreams::gzip_compressor());
	out.push(outStream);

	//Copy data through stream
	boost::iostreams::copy(inStream, out);

	//Check if copy has been done correctly
	if(!boost::filesystem::is_regular_file(destPath))
		throw std::runtime_error( "Copy failure: storage path \"" +
			destPath.string() + "\" is not a regular file");
}

}   //namespace
